package ea;

import java.util.HashMap;
import java.util.Map;

import cbr.Caze;
import cbr.Decision;
import cbr.SimilarityFunction;

public class EASimilarityFunction implements SimilarityFunction {

	// ------------------------------------------------------------------
	// Use these attributes for finding attribute weights for CBR-model 1
	// ------------------------------------------------------------------
	public static final String[] ATTRIBUTES = { "Wind speed Fokstugu", "Wind direction Fokstugu", "Friction Fokstugu",
			"Temperature Fokstugu", "Wind speed Avsjoen", "Wind direction Avsjoen", "Friction Avsjoen",
			"Temperature Avsjoen", "Wind speed Hjerkinn", "Wind direction Hjerkinn", "Temperature Hjerkinn" };

	// ------------------------------------------------------------------
	// Use these attributes for finding attribute weights for CBR-model 2
	// ------------------------------------------------------------------
//	public static final String[] ATTRIBUTES = { "Wind speed Fokstugu", "Wind direction Fokstugu", "Friction Fokstugu",
//			"Temperature Fokstugu", "Visibility Fokstugu", "Speed limit Fokstugu", "Wind speed Avsjoen",
//			"Wind direction Avsjoen", "Friction Avsjoen", "Temperature Avsjoen", "Visibility Avsjoen",
//			"Speed limit Avsjoen", "Wind speed Hjerkinn", "Wind direction Hjerkinn", "Temperature Hjerkinn",
//			"Visibility Hjerkinn" };

	private static final Map<String, Double> ROAD_DIRECTIONS;

	static {
		ROAD_DIRECTIONS = new HashMap<>();
		ROAD_DIRECTIONS.put("Wind direction Avsjoen", 60.0);
		ROAD_DIRECTIONS.put("Wind direction Fokstugu", 45.0);
		ROAD_DIRECTIONS.put("Wind direction Hjerkinn", 0.0);
	}

	private Map<String, Double> weights = new HashMap<>();

	public EASimilarityFunction(ByteArrayIndividual eaIndividual) {
		String[] attributes = ATTRIBUTES;
		int[] weights = eaIndividual.getByteArray();
		int totalWeight = 0;
		for (int i = 0; i < attributes.length; i++) {
			totalWeight += weights[i];
		}
		for (int i = 0; i < attributes.length; i++) {
			this.weights.put(attributes[i], (double) weights[i] / (double) totalWeight);
		}
	}

	public static int nofRequiredBytes() {
		return ATTRIBUTES.length;
	}

	@Override
	public double calculateSimilarity(Caze caze, Caze query) {
		double similarity = 0;
		String[] attributes = ATTRIBUTES;
		for (String attribute : attributes) {
			double weight = weights.get(attribute);
			Double cazeValue = caze.getAttributeValue(attribute);
			Decision cazeDecision = caze.getDecision();
			Double queryValue = query.getAttributeValue(attribute);
			if (attribute.startsWith("Wind speed")) {
				similarity += weight * windSpeedSimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Wind direction")) {
				double roadDirection = ROAD_DIRECTIONS.get(attribute);
				similarity += weight * windDirectionSimilarity(cazeValue, cazeDecision, queryValue, roadDirection);
			} else if (attribute.startsWith("Friction")) {
				similarity += weight * frictionSimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Visibility")) {
				similarity += weight * visibilitySimilarity(cazeValue, cazeDecision, queryValue);
			} else if (attribute.startsWith("Temperature")) {
				similarity += weight * temperatureSimilarity(cazeValue, queryValue);
			} else if (attribute.startsWith("Speed limit")) {
				similarity += weight * speedLimitSimilarity(cazeValue, cazeDecision, queryValue);
			} else {
				throw new RuntimeException("Unknown attribute: " + attribute);
			}
		}
		return similarity;
	}

	private double windSpeedSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = 1.0 / (1.0 + Math.exp(-(cazeValue - 12.5) / 5.0));
		double fq = 1.0 / (1.0 + Math.exp(-(queryValue - 12.5) / 5.0));
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - 2.0 * (fq - fc);
			break;
		case CLOSED:
			similarity = 1.0 - 2.0 * (fc - fq);
			break;
		default:
			similarity = 1.0 - 2.0 * Math.abs(fq - fc);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double windDirectionSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue,
			double roadDirection) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = roadRelativeAngle(cazeValue, roadDirection);
		double fq = roadRelativeAngle(queryValue, roadDirection);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fq - fc) / 90.0;
			break;
		case CLOSED:
			similarity = 1.0 - (fc - fq) / 90.0;
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / 90.0;
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double frictionSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = Math.log(cazeValue);
		double fq = Math.log(queryValue);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / Math.log(4.0);
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / Math.log(4.0);
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / Math.log(4.0);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double visibilitySimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = Math.log(cazeValue);
		double fq = Math.log(queryValue);
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / Math.log(4.0);
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / Math.log(4.0);
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / Math.log(4.0);
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double temperatureSimilarity(Double cazeValue, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = 1.0 / (1.0 + Math.exp(-cazeValue / 2.0));
		double fq = 1.0 / (1.0 + Math.exp(-queryValue / 2.0));
		double similarity = 1.0 - 2.0 * Math.abs(fq - fc);
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double speedLimitSimilarity(Double cazeValue, Decision cazeDecision, Double queryValue) {
		if (cazeValue == null || queryValue == null) {
			return 0.0;
		}
		double fc = cazeValue;
		double fq = queryValue;
		double similarity;
		switch (cazeDecision) {
		case OPEN:
			similarity = 1.0 - (fc - fq) / 40.0;
			break;
		case CLOSED:
			similarity = 1.0 - (fq - fc) / 40.0;
			break;
		default:
			similarity = 1.0 - Math.abs(fq - fc) / 40.0;
			break;
		}
		return Math.max(0.0, Math.min(1.0, similarity));
	}

	private double roadRelativeAngle(double direction, double roadDirection) {
		double angle = direction - roadDirection;
		angle = ((angle % 180.0) + 180.0) % 180.0;
		angle = Math.min(angle, 180.0 - angle);
		return angle;
	}
}
